import os
import sys
from pathlib import Path
from typing import TypedDict, cast
from contextlib import contextmanager
import tomlkit

from zet.entities import Stop

ZET_APP_NAME = "zet"
ZET_CACHE_DB_NAME = "cache.db"
ZET_SETTINGS_FILE_NAME = "settings.toml"
ZET_STYLESHEET_FILE_NAME = "styles.tcss"


class Config(TypedDict):
    favourite_stops: list[str]


def get_config_dir() -> Path:
    """Returns the path to zet config directory"""
    path = _get_config_dir()
    path.mkdir(parents=True, exist_ok=True)
    return path


def _get_config_dir() -> Path:
    # On Windows, store the config in roaming appdata
    if sys.platform == "win32" and "APPDATA" in os.environ:
        return Path(os.getenv("APPDATA")) / ZET_APP_NAME

    # Respect XDG_CONFIG_HOME env variable if set
    # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    if "XDG_CONFIG_HOME" in os.environ:
        config_home = Path(os.environ["XDG_CONFIG_HOME"]).expanduser()
        return config_home / ZET_APP_NAME

    # Default to ~/.config/zet/
    return Path.home() / ".config" / ZET_APP_NAME


def get_cache_dir() -> Path:
    """Returns the path to the zet cache directory"""
    path = _get_cache_dir()
    path.mkdir(parents=True, exist_ok=True)
    return path


def _get_cache_dir() -> Path:
    # Windows
    if sys.platform == "win32" and "LOCALAPPDATA" in os.environ:
        return Path(os.environ["LOCALAPPDATA"], ZET_APP_NAME)

    # Mac OS
    if sys.platform == "darwin":
        return Path.home() / "Library" / "Caches" / ZET_APP_NAME

    # Respect XDG_CONFIG_HOME env variable if set
    # https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
    if "XDG_CACHE_HOME" in os.environ:
        return Path(os.environ["XDG_CACHE_HOME"], ZET_APP_NAME)

    return Path.home() / ".cache" / ZET_APP_NAME


def get_cache_db_path() -> Path:
    return get_cache_dir() / ZET_CACHE_DB_NAME


def get_settings_path() -> Path:
    return get_config_dir() / ZET_SETTINGS_FILE_NAME


def get_stylesheet_path() -> Path:
    return get_config_dir() / ZET_STYLESHEET_FILE_NAME


def empty_config() -> Config:
    return {
        "favourite_stops": [],
    }


def load_config() -> Config:
    path = get_settings_path()
    path.parent.mkdir(parents=True, exist_ok=True)

    if path.exists():
        with path.open("r") as f:
            config = tomlkit.load(f)
            return cast(Config, config)
    else:
        return empty_config()


def save_config(config: Config):
    path = get_settings_path()
    with path.open("w") as f:
        tomlkit.dump(config, f)


@contextmanager
def edit_config():
    config = load_config()
    try:
        yield config
    finally:
        save_config(config)


def add_favourite_stop(stop: Stop):
    with edit_config() as config:
        favourite_stops = config.get("favourite_stops", [])
        favourite_stops.append(stop["id"])
        config["favourite_stops"] = list(set(favourite_stops))


def remove_favourite_stop(stop: Stop):
    with edit_config() as config:
        favourite_stops = config.get("favourite_stops", [])
        favourite_stops = [s for s in favourite_stops if s != stop["id"]]
        config["favourite_stops"] = favourite_stops



import logging
import math
import time
from contextlib import contextmanager
from dataclasses import dataclass
from functools import lru_cache
from math import sqrt

from zet import database
from zet.database import Point, get_all_lines
from zet.realtime import VehiclePosition

logger = logging.getLogger(__name__)

# Minimal distance of a point to line to be drawn when drawing with braille
MIN_DISTANCE = 0.5

# Ratio is the amount by which longitude and latitude are multiplied to get x and y
MIN_RATIO = 1000
MAX_RATIO = 20_000

# Zoom is one of 12 discrete levels which map to [MIN_RATIO, MAX_RATIO] range
MIN_ZOOM = 1
MAX_ZOOM = 12
DEFAULT_ZOOM = 5

# Zoom to ratio mapping is exponential which feels nicer than linear
_ZOOM_FACTOR = (MAX_RATIO / MIN_RATIO) ** (1 / (MAX_ZOOM - MIN_ZOOM))


def zoom_to_ratio(zoom: int) -> float:
    zoom = min(max(zoom, MIN_ZOOM), MAX_ZOOM)
    return MIN_RATIO * _ZOOM_FACTOR ** (zoom - MIN_ZOOM)


@dataclass
class Bounds:
    center_lat: float
    center_lon: float
    min_lat: float
    max_lat: float
    min_lon: float
    max_lon: float

    @property
    def dlat(self):
        return self.max_lat - self.min_lat

    @property
    def dlon(self):
        return self.max_lon - self.min_lon


@dataclass(eq=True, frozen=True)
class MapSize:
    zoom: int
    ratio: float
    char_width: int
    char_height: int
    dot_width: int
    dot_height: int
    center_lat: float
    center_lon: float
    min_lat: float
    min_lon: float
    max_lat: float
    max_lon: float

    @staticmethod
    def new(char_width: int, char_height: int, center_lat: float, center_lon: float, zoom: int):
        dot_width = char_width * 2
        dot_height = char_height * 4

        ratio = zoom_to_ratio(zoom)
        dlon = dot_width / ratio
        dlat = dot_height / ratio

        min_lat = center_lat - dlat / 2
        min_lon = center_lon - dlon / 2
        max_lat = center_lat + dlat / 2
        max_lon = center_lon + dlon / 2

        return MapSize(
            zoom=zoom,
            ratio=ratio,
            char_width=char_width,
            char_height=char_height,
            dot_width=dot_width,
            dot_height=dot_height,
            center_lat=center_lat,
            center_lon=center_lon,
            min_lat=min_lat,
            min_lon=min_lon,
            max_lat=max_lat,
            max_lon=max_lon,
        )


class Image:
    def __init__(self, width: int, height: int):
        assert width % 2 == 0
        assert height % 4 == 0

        self.width = width
        self.height = height
        self.pixels = bytearray(width * height)

    def draw_line(self, x1: int, y1: int, x2: int, y2: int):
        if x1 == x2 and y1 == y2:
            return

        x_min = min(x1, x2)
        x_max = max(x1, x2)
        x_range = range(x_min, x_max + 1)

        y_min = min(y1, y2)
        y_max = max(y1, y2)
        y_range = range(y_min, y_max + 1)

        for y in y_range:
            for x in x_range:
                if x >= 0 and x < self.width and y >= 0 and y < self.height:
                    if distance(x, y, x1, y1, x2, y2) <= MIN_DISTANCE:
                        self.pixels[y * self.width + x] = 1

    def string(self):
        string = ""
        for y in range(0, self.height, 4):
            for x in range(0, self.width, 2):
                char = self._get_char_at(x, y)
                string += char
            string += "\n"
        return string

    def to_char_list(self):
        for y in range(0, self.height, 4):
            for x in range(0, self.width, 2):
                yield self._get_char_at(x, y)

    def dump(self):
        print(self.string())

    def _get_char_at(self, x, y):
        """
        https://en.wikipedia.org/wiki/Braille_Patterns
        """
        p1 = self.pixels[(y + 0) * self.width + x]
        p2 = self.pixels[(y + 1) * self.width + x]
        p3 = self.pixels[(y + 2) * self.width + x]
        p7 = self.pixels[(y + 3) * self.width + x]
        p4 = self.pixels[(y + 0) * self.width + x + 1]
        p5 = self.pixels[(y + 1) * self.width + x + 1]
        p6 = self.pixels[(y + 2) * self.width + x + 1]
        p8 = self.pixels[(y + 3) * self.width + x + 1]

        n = p8 * 128 + p7 * 64 + p6 * 32 + p5 * 16 + p4 * 8 + p3 * 4 + p2 * 2 + p1

        return chr(10240 + n)


@contextmanager
def perflog(label: str):
    start = time.monotonic()
    yield
    duration = time.monotonic() - start
    logger.info(f"{label} took {duration:.3f}s")


@perflog("draw_map")
@lru_cache
def draw_map(map_size: MapSize) -> list[str]:
    image = Image(map_size.dot_width, map_size.dot_height)
    lines = database.get_lines(
        min_lat=map_size.min_lat,
        min_lon=map_size.min_lon,
        max_lat=map_size.max_lat,
        max_lon=map_size.max_lon,
    )

    def project(lat: float, lon: float):
        x = round((lon - map_size.min_lon) * map_size.ratio)
        y = round((lat - map_size.min_lat) * map_size.ratio)
        return x, map_size.dot_height - y

    for [(lat1, lon1), (lat2, lon2)] in lines:
        x1, y1 = project(lat1, lon1)
        x2, y2 = project(lat2, lon2)
        image.draw_line(x1, y1, x2, y2)

    return list(image.to_char_list())


def draw_full_map(zoom: int):
    ratio = zoom_to_ratio(zoom)
    lines = get_all_lines()
    bounds = get_bounds(lines)

    min_x = math.floor(bounds.min_lon * ratio)
    max_x = math.ceil(bounds.max_lon * ratio)
    min_y = math.floor(bounds.min_lat * ratio)
    max_y = math.ceil(bounds.max_lat * ratio)

    dot_width = max_x - min_x
    dot_height = max_y - min_y

    # Align for braille
    if dot_width % 2 != 0:
        dot_width += 2 - dot_width % 2

    if dot_height % 4 != 0:
        dot_height += 4 - dot_height % 4

    char_width = dot_width // 2
    char_height = dot_height // 2

    map_size = MapSize(
        zoom=zoom,
        ratio=ratio,
        char_width=char_width,
        char_height=char_height,
        dot_width=dot_width,
        dot_height=dot_height,
        center_lat=bounds.center_lat,
        center_lon=bounds.center_lon,
        min_lat=bounds.min_lat,
        min_lon=bounds.min_lon,
        max_lat=bounds.max_lat,
        max_lon=bounds.max_lon,
    )

    return draw_map(map_size)


def place_vehicle_labels(
    bounds: MapSize,
    vehicles: list[VehiclePosition],
) -> dict[tuple[int, int], str]:
    labels: dict[tuple[int, int], str] = {}

    for vehicle in vehicles:
        x = round((vehicle.longitude - bounds.min_lon) * bounds.ratio / 2)
        y = round((vehicle.latitude - bounds.min_lat) * bounds.ratio / 4)

        if x < 0 or y < 0 or x >= bounds.char_width or y >= bounds.char_height:
            continue

        y = bounds.char_height - y
        label = vehicle.route_id

        # If needed, move label left so there's room to place the full label
        min_x = bounds.char_width - len(label) - 1
        if x > min_x:
            x = min_x

        # Avoid overlap with other vehicles by moving along the y axis up or down
        placed = False
        for dy in [0, 1, -1, 2, -2, 3, -3, 4, -4]:
            if y + dy < bounds.char_height:
                # Check overlaps for the lenght of the label and the space after the
                # label so we don't print 1314 for trams 13 and 14 next to each other
                # TODO: this does not always work correctly?
                points = [(x + dx, y + dy) for dx in range(len(label) + 1)]
                if not _any_overlaps(labels, points):
                    for dx, c in enumerate(label):
                        labels[x + dx, y + dy] = c
                    placed = True
                    break

        if not placed:
            # TODO: what if we can't place it?
            logger.warning(f"Cannot place label: '{label}' at {x}:{y}")

    return labels


def _any_overlaps(labels: dict[tuple[int, int], str], points: list[tuple[int, int]]) -> bool:
    for point in points:
        if point in labels:
            return True

    return False


def get_bounds(lines: list[tuple[Point, Point]]) -> Bounds:
    count = 0
    sum_lat = 0
    sum_lon = 0
    min_lat = 180
    min_lon = 180
    max_lat = -180
    max_lon = -180

    for line in lines:
        for lat, lon in line:
            if lat < min_lat:
                min_lat = lat
            if lat > max_lat:
                max_lat = lat
            if lon < min_lon:
                min_lon = lon
            if lon > max_lon:
                max_lon = lon
            sum_lat += lat
            sum_lon += lon
            count += 1

    return Bounds(
        center_lat=sum_lat / count,
        center_lon=sum_lon / count,
        max_lat=max_lat,
        max_lon=max_lon,
        min_lat=min_lat,
        min_lon=min_lon,
    )


def distance(x0: float, y0: float, x1: float, y1: float, x2: float, y2: float) -> float:
    """
    Returns the distance of point (x0, y0) from the line that passes through
    points (x1, y1) and (x2, y2).
    https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line#Line_defined_by_two_points
    """
    dx = x2 - x1
    dy = y2 - y1
    return abs(dy * x0 - dx * y0 + x2 * y1 - y2 * x1) / sqrt(dy**2 + dx**2)


if __name__ == "__main__":
    draw_full_map(12)



from datetime import datetime

from textual import on, work
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import DataTable, Input, Label, Static, TabPane

from zet import api, config
from zet.entities import IncomingTrip, Stop
from zet.output import format_time
from zet.tui.widgets import Title


class StopsPane(TabPane):
    BINDINGS = [
        Binding("f", "toggle_favourite", "Toggle Favourite"),
    ]

    DEFAULT_CSS = """
    #stops_search {
        margin-bottom: 1;
    }
    #stops_list {
        width: 1fr;
    }
    #stops_details {
        width: 1fr;
        background: $surface;
        border-left: thick $secondary;
    }
    """

    def __init__(self, stops: list[Stop]) -> None:
        super().__init__("Stops", id="stops_pane")
        self.stops = stops
        self.stops_map = {s["id"]: s for s in stops}
        self.selected_stop: Stop | None = None
        self.favourite_stop_ids = set(config.load_config()["favourite_stops"])

    def compose(self):
        yield Vertical(
            Input(id="stops_search", placeholder="Search stops"),
            Horizontal(
                Vertical(
                    Title("Stops"),
                    DataTable(cursor_type="row", id="stops_list"),
                ),
                StopDetails(id="stops_details"),
            ),
        )

    def on_mount(self):
        dt = self.query_one(DataTable)
        dt.add_column("F", key="favourite")
        dt.add_column("ID", key="id")
        dt.add_column("Name", key="name")
        dt.add_column("Routes", key="routes")

        for stop in self._get_filtered_stops():
            self.add_stop_row(dt, stop)

    @on(Input.Changed, "#stops_search")
    def on_search(self, message: Input.Changed):
        query = message.value.strip().lower()
        dt = self.query_one(DataTable)
        dt.clear()
        for stop in self._get_filtered_stops(query):
            self.add_stop_row(dt, stop)

    def _get_filtered_stops(self, query: str = ""):
        def sort_key(stop: Stop):
            # This places favourited stops on top
            return stop["id"] not in self.favourite_stop_ids

        query = query.strip()
        if query:
            return sorted(
                (s for s in self.stops if query in s["normalizedSearchName"]),
                key=sort_key,
            )
        else:
            return sorted(self.stops, key=sort_key)

    @on(DataTable.RowHighlighted, "#stops_list")
    def on_stop_selected(self, message: DataTable.RowHighlighted):
        assert message.row_key.value is not None
        stop = self.stops_map[message.row_key.value]
        self.selected_stop = stop
        self.query_one(StopDetails).stop = stop

    def add_stop_row(self, dt: DataTable, stop: Stop):
        favourite = "*" if stop["id"] in self.favourite_stop_ids else ""
        routes = ", ".join([t["routeCode"] for t in stop["trips"]])
        dt.add_row(
            favourite,
            stop["id"],
            stop["name"],
            routes,
            key=stop["id"],
        )

    def action_toggle_favourite(self):
        if stop := self.selected_stop:
            if stop["id"] in self.favourite_stop_ids:
                config.remove_favourite_stop(stop)
            else:
                config.add_favourite_stop(stop)

            self.favourite_stop_ids = set(config.load_config()["favourite_stops"])
            is_favourite = stop["id"] in self.favourite_stop_ids

            # Update table
            dt = self.query_one(DataTable)
            dt.update_cell(stop["id"], "favourite", "*" if is_favourite else "")

            # Re-sort the table and reselect the row
            self._sort_by_favourite(dt)
            dt.move_cursor(row=dt.get_row_index(stop["id"]))

    def _sort_by_favourite(self, dt: DataTable):
        dt = self.query_one(DataTable)
        dt.sort("favourite", key=lambda f: f == "")


class StopDetails(Widget):
    stop: reactive[Stop | None] = reactive(None, recompose=True)

    def compose(self):
        if self.stop is not None:
            yield Title(f"{self.stop['name']} ({self.stop['id']})")
            yield Vertical(id="stop_trips")
            self.load_trips(self.stop["id"])
        else:
            yield Label("No stop selected")

    @work(exclusive=True)
    async def load_trips(self, stop_id: str):
        # TODO: handle exceptions
        response = await api.get_incoming_trips(self.app.session, stop_id)  # type: ignore
        trips = response.json()
        widget = IncomingTrips(trips)
        self.query_one("#stop_trips", Vertical).mount(widget)


class IncomingTrips(Widget):
    def __init__(self, trips: list[IncomingTrip]) -> None:
        super().__init__()
        self.trips = trips

    def compose(self):
        dt = DataTable(cursor_type="row", id="incoming_trips_table")
        dt.add_columns("Route", "Headsign", "Arrival", "Tracked?")
        for trip in self.trips:
            dt.add_row(
                trip["routeShortName"],
                trip["headsign"],
                format_time(trip["expectedArrivalDateTime"]),
                "[green]Yes[/]" if trip["hasLiveTracking"] else "[gray]No[/]",
            )

        yield dt

        # TODO: make this update as time passes
        now = datetime.now()
        yield Static()
        yield Static(f"Fetched: {now}")



"""
Extract tram routes from OSM and dump them into GeoJson.
"""

import json
from pathlib import Path

import duckdb


ROOT_DIR = Path(__file__).parent


def main():
    pbf = ROOT_DIR / "croatia-latest.osm.pbf"

    duckdb.install_extension("spatial")
    duckdb.load_extension("spatial")

    print("Loading relations...")
    relations = duckdb.query(f"""
        SELECT tags, refs, ref_roles, ref_types
        FROM ST_ReadOSM('{pbf}')
        WHERE kind='relation'
          AND tags.route = 'tram'
          AND tags.operator = 'ZET'
    """).fetchall()

    print("Loading ways...")
    way_ids = flatten(get_way_ids(r) for r in relations)
    ways = duckdb.query(f"""
        SELECT id, refs
        FROM ST_ReadOSM('{pbf}')
        WHERE kind='way'
          AND id IN {way_ids}
    """).fetchall()
    ways_map = {id: refs for id, refs in ways}

    print("Loading nodes...")
    node_ids = flatten([w[1] for w in ways])
    nodes = duckdb.query(f"""
        SELECT id, lat, lon
        FROM ST_ReadOSM('{pbf}')
        WHERE kind='node'
          AND id IN {node_ids}
    """).fetchall()
    nodes_map = {id: (lat, lon) for id, lat, lon in nodes}

    for relation in relations:
        collection = {"type": "FeatureCollection", "features": []}
        filename = f"{relation[0]['route']}_{relation[0]['ref'].rjust(2, '0')}.geojson"
        print(f"Writing {filename}")

        way_ids = get_way_ids(relation)
        for way_id in way_ids:
            coordinates = []
            node_ids = ways_map[way_id]
            for node_id in node_ids:
                lat, lon = nodes_map[node_id]
                coordinates.append([lon, lat])
            collection["features"].append(
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "LineString",
                        "coordinates": coordinates,
                    },
                }
            )
            with open(ROOT_DIR / filename, "w") as f:
                json.dump(collection, f, indent=4)


def get_way_ids(relation):
    _, refs, roles, types = relation
    for ref, role, type in zip(refs, roles, types):
        if role is None and type == "way":
            yield ref


def flatten(xss):
    return [x for xs in xss for x in xs]


if __name__ == "__main__":
    main()



import click
import logging
import sqlite3

from pathlib import Path
from sqlite3 import Connection
from typing import Any, NamedTuple

from zet.config import get_cache_db_path


logger = logging.getLogger(__name__)


def get_conn():
    path = get_cache_db_path()
    conn = sqlite3.connect(path, autocommit=False)
    conn.row_factory = sqlite3.Row  # Allows name-based access to columns
    migrate(conn)
    return conn


def select(conn: Connection, query: str, *params: Any) -> list[sqlite3.Row]:
    cursor = conn.execute(query, params)
    return cursor.fetchall()


def select_one(conn: Connection, query: str, params: list[Any]) -> sqlite3.Row:
    cursor = conn.execute(query, params)
    return cursor.fetchone()


def table_exists(conn: Connection, table_name: str):
    query = """
        SELECT count(*) AS count
        FROM sqlite_master
        WHERE type='table' AND name=?
    """
    row = select_one(conn, query, [table_name])
    return row["count"] > 0


def migrate(conn: Connection):
    migrations_dir = Path(__file__).parent / "migrations"
    applied_migrations = _get_applied_migrations(conn)
    available_migrations = [
        path for path in migrations_dir.iterdir() if path.is_file and path.suffix == ".sql"
    ]

    try:
        for path in available_migrations:
            if path.name not in applied_migrations:
                _apply_migration(conn, path)
                conn.commit()
    except Exception:
        logger.exception("Failed applying migrations")
        raise click.ClickException("Failed applying migrations")


def _get_applied_migrations(conn: Connection) -> list[str]:
    if table_exists(conn, "migrations"):
        rows = select(conn, "SELECT name FROM migrations ORDER BY name ASC")
        return [r["name"] for r in rows]
    else:
        return []


def _apply_migration(conn: Connection, path: Path):
    logger.info(f"Applying migration: {path.name}")

    with path.open("r") as f:
        sql = f.read()

    with conn:
        conn.executescript(sql)
        conn.execute(
            """
            INSERT INTO migrations(name, applied_at)
            VALUES (?, datetime('now'))
            """,
            [path.name],
        )


class Point(NamedTuple):
    lat: float
    lon: float


def get_lines(
    min_lat: float,
    min_lon: float,
    max_lat: float,
    max_lon: float,
) -> list[tuple[Point, Point]]:
    conn = get_conn()
    params = {"min_lat": min_lat, "min_lon": min_lon, "max_lat": max_lat, "max_lon": max_lon}
    cursor = conn.execute(
        """
        SELECT lat1, lon1, lat2, lon2
        FROM map_lines
        WHERE (lat1 >= :min_lat AND lat1 <= :max_lat AND lon1 >= :min_lon AND lon1 <= :max_lon)
           OR (lat2 >= :min_lat AND lat2 <= :max_lat AND lon2 >= :min_lon AND lon2 <= :max_lon)
        """,
        params,
    )
    lines = cursor.fetchall()
    conn.commit()

    return [
        (
            Point(line["lat1"], line["lon1"]),
            Point(line["lat2"], line["lon2"]),
        )
        for line in lines
    ]


def get_all_lines() -> list[tuple[Point, Point]]:
    conn = get_conn()
    cursor = conn.execute("SELECT lat1, lon1, lat2, lon2 FROM map_lines")
    lines = cursor.fetchall()
    conn.commit()

    return [
        (
            Point(line["lat1"], line["lon1"]),
            Point(line["lat2"], line["lon2"]),
        )
        for line in lines
    ]



import asyncio
from datetime import datetime
import logging

from aiohttp import ClientSession
from dateutil.parser import parse

from zet import api, config, database
from zet.http import make_session
from zet.utils import iterate_pairs

logger = logging.getLogger(__name__)


async def load_map_lines_cached(max_days=30):
    conn = database.get_conn()
    cursor = conn.execute("SELECT * FROM meta")
    meta = cursor.fetchone()
    conn.commit()

    imported_at = meta["map_lines_imported_at"]
    imported_at = parse(imported_at) if imported_at else datetime.min
    days_since = (datetime.now() - imported_at).days

    if days_since > max_days:
        logger.debug(f"map_lines loaded {days_since} days ago, updating")
        await load_map_lines()
    else:
        logger.debug(f"map_lines loaded {days_since} days ago, no need to update")


async def load_map_lines():
    conn = database.get_conn()
    conn.execute("DELETE FROM map_lines;")

    logger.info("Loading shapes...")
    async with make_session() as session:
        response = await api.get_shapes(session)
        shapes = response.json()

    logger.info(f"Processing {len(shapes)} shapes...")
    values = set()
    for name, points in shapes.items():
        for p1, p2 in iterate_pairs(points):
            values.add((p1["latitude"], p1["longitude"], p2["latitude"], p2["longitude"]))

    cursor = conn.executemany(
        "INSERT INTO map_lines(lat1, lon1, lat2, lon2) VALUES (?, ?, ?, ?)", values
    )
    logger.info(f"Imported {cursor.rowcount} map lines")
    conn.execute("UPDATE meta SET map_lines_imported_at = datetime('now')")
    conn.commit()


async def load_routes(session: ClientSession):
    logger.info("Loading routes...")

    path = _routes_path()
    response = await api.get_routes(session)
    with open(path, "w") as f:
        f.write(response.body)

    logger.info(f"Routes saved to {path}")


async def load_shapes(session: ClientSession):
    logger.info("Loading shapes...")

    path = _shapes_path()
    response = await api.get_shapes(session)
    with open(path, "w") as f:
        f.write(response.body)

    logger.info(f"Shapes saved to {path}")


async def load_stops(session: ClientSession):
    logger.info("Loading stops...")

    path = _stops_path()
    response = await api.get_stops(session)
    with open(path, "w") as f:
        f.write(response.body)

    logger.info(f"Stops saved to {path}")


async def load_all(session: ClientSession):
    asyncio.gather(
        load_routes(session),
        load_shapes(session),
        load_stops(session),
    )


def _routes_path():
    return config.get_cache_dir() / "routes.json"


def _shapes_path():
    return config.get_cache_dir() / "shapes.json"


def _stops_path():
    return config.get_cache_dir() / "stops.json"


async def _main():
    async with make_session() as session:
        await load_routes(session)
        await load_shapes(session)
        await load_stops(session)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    asyncio.run(load_map_lines_if_needed())



import asyncio
from os import path
from pathlib import Path

from aiohttp import ClientSession
from textual import work
from textual.app import App
from textual.binding import Binding
from textual.containers import Vertical
from textual.screen import Screen
from textual.types import CSSPathType
from textual.widgets import Footer, Header, Static, TabbedContent

from zet import ASCII_LOGO, api
from zet.config import get_stylesheet_path
from zet.entities import News, Route, Stop
from zet.tui.panes.map_pane import MapPane
from zet.tui.panes.news import NewsPane
from zet.tui.panes.routes import RoutesPane
from zet.tui.panes.stops import StopsPane


class ZetApp(App[None]):
    def __init__(self, session: ClientSession):
        super().__init__(css_path=self._get_css_paths())
        self.session = session
        self.animation_level = "none"

    def _get_css_paths(self) -> CSSPathType:
        base_css = Path("./app.css")
        user_css = get_stylesheet_path()
        return [base_css, user_css] if path.exists(user_css) else [base_css]

    async def on_mount(self):
        self.push_screen(LoadingScreen())
        self.load()

    @work
    async def load(self):
        stops, routes, news = await asyncio.gather(
            api.get_stops(self.session),
            api.get_routes(self.session),
            api.get_newsfeed(self.session),
        )
        self.switch_screen(MainScreen(stops.json(), routes.json(), news.json()))


class MainScreen(Screen):
    BINDINGS = [
        Binding("f1", "goto_stops", "Stops"),
        Binding("f2", "goto_routes", "Routes"),
        Binding("f3", "goto_news", "News"),
        Binding("f4", "goto_map", "Map"),
    ]

    def __init__(self, stops: list[Stop], routes: list[Route], news: list[News]) -> None:
        super().__init__()
        self.stops = stops
        self.routes = routes
        self.news = news

    def compose(self):
        with TabbedContent():
            yield StopsPane(self.stops)
            yield RoutesPane(self.routes)
            yield NewsPane(self.news)
            yield MapPane()
        yield Footer()

    def action_goto_stops(self):
        self.query_one(TabbedContent).active = "stops_pane"

    def action_goto_routes(self):
        self.query_one(TabbedContent).active = "routes_pane"

    def action_goto_news(self):
        self.query_one(TabbedContent).active = "news_pane"

    def action_goto_map(self):
        self.query_one(TabbedContent).active = "map_pane"


class LoadingScreen(Screen):
    TITLE = "Loading"

    DEFAULT_CSS = """
    LoadingScreen {
        Vertical {
            align: center middle;
            content-align: center middle;
            color: #1264ab;  /* zet blue */

            Static {
                width: auto;
            }
        }
    }
    """

    def compose(self):
        yield Header()
        yield Vertical(
            Static(ASCII_LOGO),
        )



from dateutil.parser import parse
from textual.containers import VerticalScroll
from textual.widgets import Markdown, TabPane

from zet.entities import News


class NewsPane(TabPane):
    def __init__(self, news: list[News]) -> None:
        super().__init__("News", id="news_pane")
        self.news = news

    def compose(self):
        markdown = "\n".join(self._news_generator())
        with VerticalScroll():
            yield Markdown(markdown)

    def _news_generator(self):
        first = True
        for item in self.news:
            if not first:
                yield ""
                yield "---"
                yield ""
            yield f"**{item['title']}**"
            yield ""
            yield item["description"]
            yield ""
            yield item["link"]
            yield ""
            yield parse(item["datePublished"]).strftime("%d.%m.%Y %H:%M")
            first = False



from textual.widgets import Label

class Title(Label):
    DEFAULT_CSS = """
    Title {
        width: 100%;
        background: $secondary;
        content-align: center middle;
        margin-bottom: 1;
    }
    """

    def __init__(self, title: str):
        super().__init__(title)



from itertools import islice, tee
from typing import Iterable, TypeVar

T = TypeVar("T")


def iterate_pairs(iterable: Iterable[T]) -> Iterable[tuple[T, T]]:
    """Iterate over pairs of items in a list."""
    items, nexts = tee(iterable, 2)
    nexts = islice(nexts, 1, None)
    return zip(items, nexts)



from __future__ import annotations

import logging
from datetime import datetime

from textual import events, work
from textual.binding import Binding
from textual.content import Content, Span
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import TabPane

from zet import map
from zet.cache import load_map_lines_cached
from zet.database import Point
from zet.map import MapSize, place_vehicle_labels
from zet.realtime import VehiclePosition, get_vehicle_positions


logger = logging.getLogger(__name__)


class MapPane(TabPane):
    def __init__(self) -> None:
        super().__init__("Map", id="map_pane")

    async def on_mount(self):
        await load_map_lines_cached()

    def compose(self):
        yield Map()
        yield StatusBar()

    def on_map_move(self, message: Map.Move):
        self.query_one(StatusBar).center = message.center
        self.query_one(StatusBar).zoom = message.zoom


class StatusBar(Widget):
    DEFAULT_CSS = """
    StatusBar  {
        height: 1;
        content-align: right middle;
    }
    """

    time = reactive(datetime.now())
    center: reactive[Point | None] = reactive(None)
    zoom: reactive[int | None] = reactive(None)

    def render(self):
        time = self.time.strftime("%A, %d. %B %Y %I:%M:%S%p")
        center = f"{self.center.lat:.3f}:{self.center.lon:.3f}" if self.center else None
        zoom = f"{self.zoom}x" if self.zoom else None
        return "  ".join(x for x in [center, zoom, time] if x)

    def on_mount(self):
        self.set_interval(1, self.update_time)

    def update_time(self):
        self.time = datetime.now()


class Map(Widget, can_focus=True):
    DEFAULT_CSS = """
    Map {
        height: 1fr;
    }
    """

    ALLOW_SELECT = False

    BINDINGS = [
        Binding("left", "move_left", "Left"),
        Binding("right", "move_right", "Right"),
        Binding("up", "move_up", "Up"),
        Binding("down", "move_down", "Down"),
        Binding("plus", "zoom_in", "Zoom In"),
        Binding("minus", "zoom_out", "Zoom Out"),
    ]

    class Move(Message):
        def __init__(self, center: Point, zoom: int):
            super().__init__()
            self.center = center
            self.zoom = zoom

    center_lat = reactive(45.8)
    center_lon = reactive(15.98)
    zoom = reactive(map.DEFAULT_ZOOM)
    vehicles: reactive[list[VehiclePosition] | None] = reactive(None)

    def __init__(self):
        super().__init__()
        self.mouse_down_at: tuple[int, int] | None = None
        self.timer = None

    def on_mount(self):
        self.post_move_message()
        self.load_vehicles()
        # Seems like realtime data is updated every 10 seconds
        self.timer = self.set_interval(10, self.load_vehicles)

    @work(exclusive=True)
    async def load_vehicles(self):
        self.vehicles = await get_vehicle_positions()

    def render(self):
        logger.info("Rendering!")
        width = self.size.width
        height = self.size.height

        if width == 0 or height == 0:
            return ""

        center = Point(self.center_lat, self.center_lon)
        bounds = MapSize.new(width, height, center.lat, center.lon, self.zoom)
        image = map.draw_map(bounds)
        labels = place_vehicle_labels(bounds, self.vehicles or [])

        string = ""
        spans = []
        for y in range(bounds.char_height):
            for x in range(bounds.char_width):
                offset = y * bounds.char_width + x
                if (x, y) in labels:
                    string += labels[x, y]
                    # TODO: we could reduce the number of spans by having one per
                    # label instead of one per character
                    spans.append(Span(len(string) - 1, len(string), style="bold"))
                else:
                    string += image[offset]
            string += "\n"

        return Content(string, spans)

    def post_move_message(self):
        self.post_message(self.Move(Point(self.center_lat, self.center_lon), self.zoom))

    def action_move_left(self):
        self.center_lon -= self.move_delta
        self.post_move_message()

    def action_move_right(self):
        self.center_lon += self.move_delta
        self.post_move_message()

    def action_move_up(self):
        self.center_lat += self.move_delta
        self.post_move_message()

    def action_move_down(self):
        self.center_lat -= self.move_delta
        self.post_move_message()

    def action_zoom_in(self):
        if self.zoom < map.MAX_ZOOM:
            self.zoom += 1
        self.post_move_message()

    def action_zoom_out(self):
        if self.zoom > map.MIN_ZOOM:
            self.zoom -= 1
        self.post_move_message()

    @property
    def ratio(self):
        return map.zoom_to_ratio(self.zoom)

    @property
    def move_delta(self):
        return 12 / self.ratio

    def on_mouse_down(self, event: events.MouseDown):
        self.mouse_down_at = (event.x, event.y)

    def on_mouse_up(self, event: events.MouseUp):
        self.mouse_down_at = None

    def on_mouse_move(self, event: events.MouseMove):
        if self.mouse_down_at is not None:
            self.center_lon -= event.delta_x * 2 / self.ratio
            self.center_lat += event.delta_y * 4 / self.ratio
        self.post_move_message()

    def on_mouse_scroll_up(self, event: events.MouseScrollUp):
        self.action_zoom_in()
        self.post_move_message()

    def on_mouse_scroll_down(self, event: events.MouseScrollDown):
        self.action_zoom_out()
        self.post_move_message()



import logging

from textual import on, work
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import DataTable, Input, TabPane

from zet import api
from zet.entities import Route, Trip
from zet.output import format_time
from zet.tui.widgets import Title

logger = logging.getLogger(__name__)


class RoutesPane(TabPane):
    BINDINGS = [
        Binding("f", "toggle_favourite", "Toggle Favourite"),
    ]

    DEFAULT_CSS = """
    #routes_search {
        margin-bottom: 1;
    }
    #routes_list {
        width: 1fr;
    }
    #routes_departures_a {
        border-left: thick $secondary;
        width: 1fr;
    }
    #routes_departures_b {
        border-left: thick $secondary;
        width: 1fr;
    }
    """

    def __init__(self, routes: list[Route]) -> None:
        super().__init__("Routes", id="routes_pane")
        self.routes = routes
        self.routes_map = {s["id"]: s for s in routes}
        self.selected_route: Route | None = None

    def compose(self):
        yield Vertical(
            Input(id="routes_search", placeholder="Search routes"),
            Horizontal(
                Vertical(
                    Title("Routes"),
                    DataTable(cursor_type="row", id="routes_list"),
                ),
                RouteDepartures(title="Departures A", direction=0, id="routes_departures_a"),
                RouteDepartures(title="Departures B", direction=1, id="routes_departures_b"),
            ),
        )

    @on(Input.Changed, "#routes_search")
    def on_search(self, message: Input.Changed):
        query = message.value.strip().lower()
        dt = self.query_one("#routes_list", DataTable)
        dt.clear()
        for route in self.routes:
            if query in route["normalizedSearchName"]:
                self.add_route_row(dt, route)

    @on(DataTable.RowHighlighted, "#routes_list")
    def on_route_highlighted(self, message: DataTable.RowHighlighted):
        if message.row_key.value is not None:
            route_id = message.row_key.value
            route = self.routes_map[route_id]
            self.selected_route = route
            self.load_route_trips(route_id)

    @work(exclusive=True)
    async def load_route_trips(self, route_id: str):
        # TODO: handle exceptions
        response = await api.get_route_trips(self.app.session, route_id)  # type: ignore
        trips = response.json()
        logger.error(f"Loaded {len(trips)} trips for route {route_id}")

        self.query_one("#routes_departures_a", RouteDepartures).trips = trips
        self.query_one("#routes_departures_b", RouteDepartures).trips = trips

    def on_mount(self):
        dt = self.query_one("#routes_list", DataTable)
        dt.add_columns("ID", "From", "To")
        for route in self.routes:
            self.add_route_row(dt, route)

    def add_route_row(self, dt: DataTable, route: Route):
        # For circular trips the departure is null
        departure = (
            route["departureHeadsign"]
            if route["departureHeadsign"]
            else route["destinationHeadsign"]
        )

        dt.add_row(
            route["id"],
            departure,
            route["destinationHeadsign"],
            key=route["id"],
        )

    def action_toggle_favourite(self):
        pass  # TODO


class RouteDepartures(Widget):
    trips: reactive[list[Trip] | None] = reactive(None)

    DEFAULT_CSS = """
    DataTable {
        width: auto;
    }
    """

    def __init__(self, title: str, direction: int, id):
        self.title = title
        self.direction = direction
        super().__init__(id=id)

    def compose(self):
        dt = DataTable(cursor_type="row")
        dt.add_columns("Destination", "Departure", "Arrival", "Tracked?")
        yield Vertical(Title(self.title), dt)

    def watch_trips(self, _, new_trips: list[Trip] | None) -> None:
        dt = self.query_one(DataTable)
        dt.clear(columns=False)

        if new_trips:
            for trip in new_trips:
                if trip["direction"] == self.direction:
                    dt.add_row(
                        trip["headsign"],
                        format_time(trip["departureDateTime"]),
                        format_time(trip["arrivalDateTime"]),
                        "[green]Yes[/]" if trip["hasLiveTracking"] else "[gray]No[/]",
                        label="*" if trip["tripStatus"] == 2 else "",
                    )



"""
Fetch GTFS realtime data from ZET.

https://www.zet.hr/odredbe/datoteke-u-gtfs-formatu/669
"""

import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Generator

from aiohttp import ClientSession
from google.transit import gtfs_realtime_pb2

from zet.http import logger_trace_config


logger = logging.getLogger(__name__)


@dataclass
class VehiclePosition:
    trip_id: str
    route_id: str
    vehicle_id: str
    latitude: float
    longitude: float
    timestamp: datetime


async def get_vehicle_positions() -> list[VehiclePosition]:
    """Returns realtime vehicle positions"""
    try:
        logger.debug("Loading vehicle positions...")
        data = await fetch_realtime_data()
        positions = list(parse_realtime_data(data))
        logger.info(f"Loaded {len(positions)} vehicle positions")
        return positions
    except Exception:
        logger.exception("Failed loading vehicle positions")
        return []


def parse_realtime_data(data: bytes) -> Generator[VehiclePosition]:
    feed = gtfs_realtime_pb2.FeedMessage()  # type: ignore
    feed.ParseFromString(data)
    for entity in feed.entity:
        if entity.HasField("vehicle"):
            yield VehiclePosition(
                trip_id=entity.vehicle.trip.trip_id,
                route_id=entity.vehicle.trip.route_id,
                vehicle_id=entity.vehicle.vehicle.id,
                latitude=entity.vehicle.position.latitude,
                longitude=entity.vehicle.position.longitude,
                timestamp=datetime.fromtimestamp(entity.vehicle.timestamp),
            )


async def fetch_realtime_data() -> bytes:
    async with ClientSession(trace_configs=[logger_trace_config()]) as session:
        async with session.get("https://www.zet.hr/gtfs-rt-protobuf") as response:
            response.raise_for_status()
            return await response.read()
